from concurrent.futures import ThreadPoolExecutor
import atexit
from operator import length_hint
from tqdm import tqdm
from typing import Callable, TypeVar, Iterable


SHARED_THREAD_POOL = ThreadPoolExecutor()

def at_exit():
    SHARED_THREAD_POOL.shutdown()

atexit.register(at_exit)


U = TypeVar("U")
V = TypeVar("V")

def thread_map(
    fn: Callable[[U], V], 
    it: Iterable[U], 
    desc: str = None, 
    total: int = None, 
    unit: str = "it", 
    verbose: bool = False
) -> "list[V]":
    disable = not verbose
    total = total or length_hint(it)
    results = []
    with tqdm(desc=desc, total=total, unit=unit, disable=disable) as pbar:
        futures = SHARED_THREAD_POOL.map(fn, it)
        for result in futures:
            results.append(result)
            pbar.update()
    return results


# TODO
def thread_imap():
    ...


from .boundingbox import BoundingBox
from .annotation import Annotation
from .annotationset import AnnotationSet
from .utils import grouping, all_equal, mean
from .atomic import open_atomic

from typing import DefaultDict, Dict, Mapping, Optional, Sequence, Union, Iterable
from collections import defaultdict
import numpy as np
from copy import copy
from math import isnan
from pathlib import Path
from enum import Enum, auto
from itertools import chain, product
from functools import lru_cache
from tqdm import tqdm

class RecallSteps(Enum):
    ELEVEN = auto()
    ALL = auto()


class PartialEvaluationItem:

    def __init__(self, 
        tps: "list[bool]" = None,
        scores: "list[float]" = None,
        npos: int = 0
    ) -> None:
        self._tps = tps or []
        self._scores = scores or []
        self._npos = npos
        self._cache = {}

        assert self._npos >= 0
        assert len(self._tps) == len(self._scores)

    def __iadd__(self, other: "PartialEvaluationItem") -> "PartialEvaluationItem":
        self._tps += other._tps
        self._scores += other._scores
        self._npos += other._npos
        self.clear_cache()
        return self

    def __add__(self, other: "PartialEvaluationItem") -> "PartialEvaluationItem":
        copy_ = copy(self)
        copy_ += other
        return copy_
    
    @property
    def ndet(self) -> int:
        return len(self._tps)

    @property
    def npos(self) -> int:
        return self._npos

    def tp(self) -> int:
        tp = self._cache.get("tp", sum(self._tps))
        self._cache["tp"] = tp
        assert tp <= self._npos
        return tp

    def ap(self) -> float:
        ap = self._cache.get("ap")
        if ap is not None:
            return ap
        ap = COCOEvaluator._compute_ap(self._scores, self._tps, self._npos)
        self._cache["ap"] = ap
        return ap

    def ar(self) -> Optional[float]:
        ar = self._cache.get("ar")
        if ar is not None:
            return ar
        tp = self.tp()
        ar = tp / self._npos if self._npos != 0 else float("nan")
        self._cache["ar"] = ar
        return ar

    def clear_cache(self):
        self._cache.clear()

    def evaluate(self) -> "EvaluationItem":
        return EvaluationItem(self.tp(), self.ndet, self._npos, self.ap(), self.ar())


class EvaluationItem:
    """Evaluation of COCO metrics for one label."""

    __slots__ = ("tp", "ndet", "npos", "ap", "ar")
    
    def __init__(self, tp: int, ndet: int, npos: int, ap: Union[float, None], ar: Union[float, None]) -> None:
        self.tp = tp
        self.ndet = ndet
        self.npos = npos
        self.ap = ap
        self.ar = ar


class PartialEvaluation(DefaultDict[str, PartialEvaluationItem]):
    """Do not mutate this excepted with defined methods."""

    def __init__(self, items: Mapping[str, PartialEvaluationItem] = None):
        super().__init__(lambda: PartialEvaluationItem())
        if items is not None:
            self.update(items)
        self._cache = {}

    def __iadd__(self, other: "PartialEvaluation") -> "PartialEvaluation":
        for key, value in other.items():
            self[key] += value
        self.clear_cache()
        return self

    def __add__(self, other: "PartialEvaluation") -> "PartialEvaluation":
        copy_ = copy(self)
        copy_ += other
        return copy_

    def ap(self) -> float:
        ap = self._cache.get("ap")
        if ap is not None:
            return ap
        ap = mean(a for a in (ev.ap() for ev in self.values()) if not isnan(a))
        self._cache["ap"] = ap
        return ap

    def ar(self) -> float:
        ar = self._cache.get("ar")
        if ar is not None:
            return ar
        ar = mean(a for a in (ev.ar() for ev in self.values()) if not isnan(a))
        self._cache["ar"] = ar
        return ar

    def clear_cache(self):
        self._cache.clear()
    
    def evaluate(self) -> "Evaluation":
        return Evaluation(self)


class Evaluation(DefaultDict[str, EvaluationItem]):
    """Evaluation of COCO metrics for multiple labels."""

    def __init__(self, evaluation: PartialEvaluation) -> None:
        super().__init__(lambda: PartialEvaluation())
        self.update({label: ev.evaluate() for label, ev in evaluation.items()})
        
        self._ap = evaluation.ap()
        self._ar = evaluation.ar()

    def ap(self) -> float:
        return self._ap

    def ar(self) -> float:
        return self._ar


class MultiThresholdEvaluation(Dict[str, Dict[str, float]]):
    """Evaluation of COCO metrics for multiple labels and multiple
    IoU thresholds.
    """

    def __init__(self, evaluations: "list[Evaluation]") -> None:
        result: "defaultdict[str, list[EvaluationItem]]" = defaultdict(list)
        for evaluation in evaluations:
            for label, ev_item in evaluation.items():
                result[label].append(ev_item)

        super().__init__({
            label: {"ap": mean(ev.ap for ev in evs if not isnan(ev.ap)), "ar": mean(ev.ar for ev in evs if not isnan(ev.ar))} 
                for label, evs in result.items()})

    def ap(self) -> float:
        return mean(ev["ap"] for ev in self.values() if not isnan(ev["ap"]))
    
    def ar(self) -> float:
        return mean(ev["ar"] for ev in self.values() if not isnan(ev["ar"]))
    

class COCOEvaluator:
    """Class for evaluating standard COCO metrics efficently.
    
    This class use an internal cache to store evaluation results, 
    hence, repeated calls to '.evaluate(...)', '.ap50()' and other 
    such methods are fast.

    There are lots of asserts in the evaluation hot path to ensure
    a valid evaluation, speed gains can be otained by disabling them
    (python3 --OO ...).
    """

    AP_THRESHOLDS = np.linspace(0.5, 0.95, 10)

    SMALL_RANGE = (0.0, 32.0**2)
    MEDIUM_RANGE = (32.0**2, 96.0**2)
    LARGE_RANGE = (96.0**2, float("inf"))
    ALL_RANGE = (0.0, float("inf"))

    RECALL_STEPS = np.linspace(0.0, 1.0, 101)

    def __init__(self, 
        ground_truths: AnnotationSet,
        predictions: AnnotationSet ,
        labels: Iterable[str] = None,
    ) -> None:
        self._predictions = predictions
        self._ground_truths = ground_truths

        if labels is None:
            self.labels = []
        else:
            self.labels = list(labels)

    @lru_cache(maxsize=60 * 4)  # Enough room for 4 times all standard metrics
    def evaluate(self,
        iou_threshold: float,
        max_detections: int,
        size_range: "tuple[float, float]"
    ) -> Evaluation:
        """COCO evaluation with custom parameters. The result
        is cached so that repeated call as fast.

        Parameters:
        - iou_threshold: the bounding box iou threshold to 
        consider a ground-truth to detection association valid.
        - max_detections: the maximum number of detections taken
        into account (sorted by descreasing confidence). Defaults
        to 100.
        - size_range: the range of size (bounding box area) to 
        consider. Defaults to all sizes.

        Returns:
        - An evaluation holding the metrics.
        """
        self._assert_params(iou_threshold, max_detections, size_range)
        
        return self.evaluate_annotations(
            self._predictions, 
            self._ground_truths,
            iou_threshold, 
            max_detections, 
            size_range,
            self.labels).evaluate()

    def ap(self) -> float:
        return self.ap_evaluation().ap()

    def ap_50(self) -> float:
        return self.ap_50_evaluation().ap()

    def ap_75(self) -> float:
        return self.ap_75_evaluation().ap()

    def ap_small(self) -> float:
        return self.small_evaluation().ap()

    def ap_medium(self) -> float:
        return self.medium_evaluation().ap()
    
    def ap_large(self) -> float:
        return self.large_evaluation().ap()
    
    def ar_1(self) -> float:
        return self.ndets_1_evaluation().ar()

    def ar_10(self) -> float:
        return self.ndets_10_evaluation().ar()
    
    def ar_100(self) -> float:
        return self.ndets_100_evaluation().ar()
    
    def ar_small(self) -> float:
        return self.small_evaluation().ar()

    def ar_medium(self) -> float:
        return self.medium_evaluation().ar()

    def ar_large(self) -> float:
        return self.large_evaluation().ar()

    def ap_evaluation(self) -> MultiThresholdEvaluation:
        evaluations = [self.evaluate(t, 100, self.ALL_RANGE) for t in self.AP_THRESHOLDS]
        return MultiThresholdEvaluation(evaluations)

    def ap_50_evaluation(self) -> Evaluation:
        return self.evaluate(0.5, 100, self.ALL_RANGE)

    def ap_75_evaluation(self) -> Evaluation:
        return self.evaluate(0.75, 100, self.ALL_RANGE)

    def small_evaluation(self) -> MultiThresholdEvaluation:
        return self._range_evalation(self.SMALL_RANGE)

    def medium_evaluation(self) -> MultiThresholdEvaluation:
        return self._range_evalation(self.MEDIUM_RANGE)

    def large_evaluation(self) -> MultiThresholdEvaluation:
        return self._range_evalation(self.LARGE_RANGE)

    def ndets_1_evaluation(self) -> MultiThresholdEvaluation:
        return self._ndets_evaluation(1)

    def ndets_10_evaluation(self) -> MultiThresholdEvaluation:
        return self._ndets_evaluation(10)

    def ndets_100_evaluation(self) -> MultiThresholdEvaluation:
        return self._ndets_evaluation(100)

    def _range_evalation(self, range_: "tuple[float, float]") -> MultiThresholdEvaluation:
        evaluations = [self.evaluate(t, 100, range_) for t in self.AP_THRESHOLDS]
        return MultiThresholdEvaluation(evaluations)

    def _ndets_evaluation(self, max_dets: int) -> MultiThresholdEvaluation:
        evaluations = [self.evaluate(t, max_dets, self.ALL_RANGE) for t in self.AP_THRESHOLDS]
        return MultiThresholdEvaluation(evaluations)

    @classmethod
    def evaluate_annotations(cls,
        predictions: AnnotationSet, 
        ground_truths: AnnotationSet,
        iou_threshold: float,
        max_detections: int,
        size_range: "tuple[float, float]",
        labels: Sequence[str] = None,
    ) -> PartialEvaluation:
        image_ids = ground_truths.image_ids | predictions.image_ids
        evaluation = PartialEvaluation()

        for image_id in sorted(image_ids):  # Sorted to ensure reproductibility
            gt = ground_truths.get(image_id) or Annotation(image_id)
            pred = predictions.get(image_id) or Annotation(image_id)

            evaluation += cls.evaluate_annotation(
                pred, gt, iou_threshold, max_detections, size_range, labels)
        return evaluation

    @classmethod
    def evaluate_annotation(cls,
        prediction: Annotation, 
        ground_truth: Annotation,
        iou_threshold: float,
        max_detections: int,
        size_range: "tuple[float, float]",
        labels: Sequence[str] = None
    ) -> PartialEvaluation:
        assert prediction.image_id == ground_truth.image_id

        preds = grouping(prediction.boxes, lambda box: box.label)
        refs = grouping(ground_truth.boxes, lambda box: box.label)
        labels = labels or set(preds.keys()).union(refs.keys())
        evaluation = PartialEvaluation()

        for label in labels:
            dets = preds.get(label, [])
            gts = refs.get(label, [])
            
            evaluation[label] += cls.evaluate_boxes(
                dets, gts, iou_threshold, max_detections, size_range)

        return evaluation

    @classmethod
    def evaluate_boxes(cls,
        predictions: "list[BoundingBox]",
        ground_truths: "list[BoundingBox]",
        iou_threshold: float,
        max_detections: int,
        size_range: "tuple[float, float]"
    ) -> PartialEvaluationItem:
        assert all(p.is_detection for p in predictions)
        assert all(g.is_ground_truth for g in ground_truths)
        assert all_equal(p.label for p in predictions)
        assert all_equal(g.label for g in ground_truths)

        cls._assert_params(iou_threshold, max_detections, size_range)

        dets = sorted(predictions, key=lambda box: box._confidence, reverse=True)
        dets = dets[:max_detections]

        gts = sorted(ground_truths, key=lambda box: not box.area_in(size_range))
        gt_ignore = [not g.area_in(size_range) for g in gts]  # Redundant `area_in`
        
        gt_matches = {}
        dt_matches = {}

        for idx_dt, det in enumerate(dets):
            best_iou = 0.0
            idx_best = -1

            for idx_gt, gt in enumerate(gts):
                if idx_gt in gt_matches:
                    continue
                if idx_best > -1 and not gt_ignore[idx_best] and gt_ignore[idx_gt]:
                    break 
                
                iou = det.iou(gt)
                if iou < best_iou:
                    continue

                best_iou = iou
                idx_best = idx_gt
            
            if idx_best == -1 or best_iou < iou_threshold:
                continue

            dt_matches[idx_dt] = idx_best
            gt_matches[idx_best] = idx_dt
        
        # This is overly complicated
        dt_ignore = [gt_ignore[dt_matches[i]] if i in dt_matches else not d.area_in(size_range)
            for i, d in enumerate(dets)]

        scores = [d._confidence
            for i, d in enumerate(dets) if not dt_ignore[i]]
        matches = [i in dt_matches
            for i in range(len(dets)) if not dt_ignore[i]]
        npos = sum(1 for i in range(len(gts)) if not gt_ignore[i])

        return PartialEvaluationItem(matches, scores, npos)

    @staticmethod
    def _assert_params(
        iou_threshold: float,
        max_detections: int, 
        size_range: "tuple[float, float]"
    ):
        assert 0.0 <= iou_threshold <= 1.0
        assert max_detections >= 0
        low, high = size_range
        assert low >= 0.0 and high >= low

    def clear_cache(self):
        self.evaluate.cache_clear()

    def _evaluate_all(self, verbose: bool = False): 
        params = chain(
            product(
                self.AP_THRESHOLDS, 
                (100,),
                (self.SMALL_RANGE, self.MEDIUM_RANGE, self.LARGE_RANGE, self.ALL_RANGE)),
            product(
                self.AP_THRESHOLDS, 
                (1, 10), 
                (self.ALL_RANGE,)))

        for t, d, r in tqdm(params, desc="Evaluation", total=60, disable=not verbose):
            self.evaluate(t, d, r)

    def show_summary(self, *, verbose: bool = False):
        """Compute and show the standard COCO metrics."""
        from rich.table import Table
        from rich import print as pprint

        self._evaluate_all(verbose=verbose)

        table = Table(title="COCO Evaluation", show_footer=True)
        table.add_column("Label", footer="Total")

        metrics = {
            "AP 50:95": self.ap(), "AP 50": self.ap_50(), "AP 75": self.ap_75(), 
            "AP S": self.ap_small(), "AP M": self.ap_medium(), "AP L": self.ap_large(), 
            "AR 1": self.ar_1(), "AR 10": self.ar_10(), "AR 100": self.ar_100(), 
            "AR S": self.ar_small(), "AR M": self.ar_medium(), "AR L": self.ar_large()}

        for metric_name, metric in metrics.items():
            table.add_column(metric_name, justify="right", footer=f"{metric:.2%}")

        labels = self.labels or sorted(self.ap_evaluation().keys())

        for label in labels:
            ap = self.ap_evaluation()[label]["ap"]
            ap_50 = self.ap_50_evaluation()[label].ap
            ap_75 = self.ap_75_evaluation()[label].ap

            ap_s = self.small_evaluation()[label]["ap"]
            ap_m = self.medium_evaluation()[label]["ap"]
            ap_l = self.large_evaluation()[label]["ap"]

            ar_1 = self.ndets_1_evaluation()[label]["ap"]
            ar_10 = self.ndets_10_evaluation()[label]["ap"]
            ar_100 = self.ndets_100_evaluation()[label]["ap"]

            ar_s = self.small_evaluation()[label]["ar"]
            ar_m = self.medium_evaluation()[label]["ar"]
            ar_l = self.large_evaluation()[label]["ar"]

            table.add_row(label, 
                f"{ap:.2%}", f"{ap_50:.2%}", f"{ap_75:.2%}", 
                f"{ap_s:.2%}", f"{ap_m:.2%}", f"{ap_l:.2%}", 
                f"{ar_1:.2%}", f"{ar_10:.2%}", f"{ar_100:.2%}", 
                f"{ar_s:.2%}", f"{ar_m:.2%}", f"{ar_l:.2%}")

        table.header_style = "bold"
        table.footer_style = "bold"
        table.row_styles = ["none", "dim"]
        
        for c in table.columns[1:4]: 
            c.style = "red"
            c.header_style = "red"
            c.footer_style = "red"
        for c in table.columns[4:7]: 
            c.style = "magenta"
            c.header_style = "magenta"
            c.footer_style = "magenta"
        for c in table.columns[7:10]: 
            c.style = "blue"
            c.header_style = "blue"
            c.footer_style = "blue"
        for c in table.columns[10:13]: 
            c.style = "green"
            c.header_style = "green"
            c.footer_style = "green"

        pprint(table)

    def to_csv(self, *, verbose: bool = False) -> str:
        """Compute and show the standard COCO metrics."""
        self._evaluate_all(verbose=verbose)
        labels = self.labels or sorted(self.ap_evaluation().keys())
        headers = ("label",
            "AP 50:95", "AP 50", "AP 75", 
            "AP S", "AP M", "AP L", 
            "AR 1", "AR 10", "AR 100", 
            "AR S", "AR M", "AR L")
        rows = []

        for label in labels:
            ap = self.ap_evaluation()[label]["ap"]
            ap_50 = self.ap_50_evaluation()[label].ap
            ap_75 = self.ap_75_evaluation()[label].ap

            ap_s = self.small_evaluation()[label]["ap"]
            ap_m = self.medium_evaluation()[label]["ap"]
            ap_l = self.large_evaluation()[label]["ap"]

            ar_1 = self.ndets_1_evaluation()[label]["ap"]
            ar_10 = self.ndets_10_evaluation()[label]["ap"]
            ar_100 = self.ndets_100_evaluation()[label]["ap"]

            ar_s = self.small_evaluation()[label]["ar"]
            ar_m = self.medium_evaluation()[label]["ar"]
            ar_l = self.large_evaluation()[label]["ar"]

            rows.append(",".join((label, 
                f"{ap}", f"{ap_50}", f"{ap_75}", 
                f"{ap_s}", f"{ap_m}", f"{ap_l}", 
                f"{ar_1}", f"{ar_10}", f"{ar_100}", 
                f"{ar_s}", f"{ar_m}", f"{ar_l}")))

        return "\n".join((",".join(headers), *rows))

    def save_csv(self, path: Path, *, verbose: bool = False):
        csv_data = self.to_csv(verbose=verbose)
        with open_atomic(path, "w") as f:
            f.write(csv_data)

    @classmethod
    def _compute_ap(cls, scores: "list[float]", matched: "list[bool]", NP: int) -> float:
        """ This curve tracing method has some quirks that do not appear when only unique confidence thresholds
        are used (i.e. Scikit-learn's implementation), however, in order to be consistent, the COCO's method is reproduced. 
        
        Copyrights: https://github.com/rafaelpadilla/review_object_detection_metrics
        """
        if NP == 0:
            return float("nan")

        # sort in descending score order
        scores = np.array(scores, dtype=float)
        matched = np.array(matched, dtype=bool)
        inds = np.argsort(-scores, kind="stable")

        scores = scores[inds]
        matched = matched[inds]

        tp = np.cumsum(matched)

        rc = tp / NP
        pr = tp / np.arange(1, matched.size + 1)
        # make precision monotonically decreasing
        i_pr = np.maximum.accumulate(pr[::-1])[::-1]
        rec_idx = np.searchsorted(rc, cls.RECALL_STEPS, side="left")

        sum_ = i_pr[rec_idx[rec_idx < i_pr.size]].sum()

        return sum_ / rec_idx.size


from .errors import ParsingError

from enum import Enum, auto
from typing import Mapping, Union, Tuple
import xml.etree.ElementTree as et


Coordinates = Tuple[float, float, float, float]


class BoxFormat(Enum):
    """
    The bounding box coordinate system.

    LTRB: xmin, ymin, xmax, ymax
    LTWH: xmin, ymin, width, height
    XYWH: xmid, ymid, width, height
    """
    LTRB = auto()
    LTWH = auto()
    XYWH = auto()

    @classmethod
    def from_string(cls, string: str) -> "BoxFormat":
        if string == "ltrb":
            return BoxFormat.LTRB
        elif string == "ltwh":
            return BoxFormat.LTWH
        elif string == "xywh":
            return BoxFormat.XYWH
        else:
            raise ValueError(f"Invalid BoxFormat string '{string}'")


class BoundingBox:
    """Represents a bounding box with a label and an optional confidence score.

    The bounding box coordinates are specified by the top-left corner (xmin, ymin)
    and the bottom-right corner (xmax, ymax). Coordinates are absolute (i.e. in pixels).
    
    ```
        xmin   xmid   xmax
    ymin ╆╍╍╍╍╍╍┿╍╍╍╍╍╍┪
         ╏      ┆      ╏
    ymid ╂┄┄┄┄┄┄┼┄┄┄┄┄┄┨
         ╏      ┆      ╏
    ymax ┺╍╍╍╍╍╍┴╍╍╍╍╍╍┛
    ```
    
    Use the '.create(...)' classmethod to create a BoundingBox using a different coordinate
    system."""

    __slots__ = ("label", "_xmin", "_ymin", "_xmax", "_ymax", "_confidence")

    def __init__(self,
        label: str, 
        xmin: float, 
        ymin: float, 
        xmax: float,
        ymax: float, *,
        confidence: float = None
    ) -> None:
        assert xmin <= xmax, "`xmax` must be greater than `xmin`."
        assert ymin <= ymax, "`ymax` must be greater than `ymin`."

        if confidence: 
            assert 0.0 <= confidence <= 1.0, \
                f"Confidence ({confidence}) should be in [0, 1]."

        self.label = label
        self._xmin = xmin
        self._ymin = ymin
        self._xmax = xmax
        self._ymax = ymax
        self._confidence = confidence

    @property
    def confidence(self) -> float:
        return self._confidence

    @confidence.setter
    def confidence(self, confidence: float):
        assert 0.0 <= confidence <= 1.0, \
                f"Confidence ({confidence}) should be in [0, 1]."
        self._confidence = confidence

    @property
    def xmin(self) -> float:
        return self._xmin

    @property
    def ymin(self) -> float:
        return self._ymin

    @property
    def xmax(self) -> float:
        return self._xmax

    @property
    def ymax(self) -> float:
        return self._ymax

    @property
    def xmid(self) -> float:
        return (self._xmin + self._xmax) / 2.0

    @property
    def ymid(self) -> float: 
        return (self._ymin + self._ymax) / 2.0
    
    @property
    def width(self) -> float: 
        return self._xmax - self._xmin

    @property
    def height(self) -> float: 
        return self._ymax - self._ymin

    @property
    def area(self) -> float:
        return self.width * self.height

    def area_in(self, range_: "tuple[float, float]") -> bool:
        lower_bound, upper_bound = range_
        return lower_bound <= self.area <= upper_bound

    @property
    def pascal_area(self) -> int:
        width = int(self._xmax) - int(self._xmin) + 1
        height = int(self._ymax) - int(self._ymin) + 1
        return width * height

    def iou(self, other: "BoundingBox") -> float:
        """Intersection over Union computed with float coordinates."""
        xmin = max(self._xmin, other._xmin)
        ymin = max(self._ymin, other._ymin)
        xmax = min(self._xmax, other._xmax)
        ymax = min(self._ymax, other._ymax)

        if xmax < xmin or ymax < ymin:
            return 0.0

        intersection = (xmax - xmin) * (ymax - ymin)
        union = self.area + other.area - intersection

        if union == 0.0:
            return 1.0

        return intersection / union

    def pascal_iou(self, other: "BoundingBox") -> float:
        """Intersection over Union computed with integer coordinates."""
        xmin = max(int(self._xmin), int(other._xmin))
        ymin = max(int(self._ymin), int(other._ymin))
        xmax = min(int(self._xmax), int(other._xmax))
        ymax = min(int(self._ymax), int(other._ymax))

        if xmax < xmin or ymax < ymin:
            return 0.0

        intersection = (xmax - xmin + 1) * (ymax - ymin + 1)
        union = self.pascal_area + other.pascal_area - intersection

        if union == 0:
            return 1.0

        return intersection / union

    @property
    def is_detection(self) -> bool:
        return self._confidence is not None

    @property
    def is_ground_truth(self) -> bool:
        return self._confidence is None

    @staticmethod
    def rel_to_abs(coords: Coordinates, size: "tuple[int, int]") -> Coordinates:
        a, b, c, d = coords
        w, h = size
        return a*w, b*h, c*w, d*h
    
    @staticmethod
    def abs_to_rel(coords: Coordinates, size: "tuple[int, int]") -> Coordinates:
        a, b, c, d = coords
        w, h = size
        return a/w, b/h, c/w, d/h

    @staticmethod
    def ltwh_to_ltrb(coords: Coordinates) -> Coordinates:
        xmin, ymin, width, height = coords
        return xmin, ymin, xmin + width, ymin + height

    @staticmethod
    def xywh_to_ltrb(coords: Coordinates) -> Coordinates:
        xmid, ymid, width, height = coords
        w_h, h_h = width/2, height/2
        return xmid-w_h, ymid-h_h, xmid+w_h, ymid+h_h

    @property
    def ltrb(self) -> Coordinates:
        return self._xmin, self._ymin, self._xmax, self._ymax

    @property
    def ltwh(self) -> Coordinates:
        return self._xmin, self._ymin, self.width, self.height

    @property
    def xywh(self) -> Coordinates:
        return self.xmid, self.ymid, self.width, self.height

    @classmethod
    def create(cls, *,
        label: str, 
        coords: Coordinates, 
        confidence: float = None,
        box_format = BoxFormat.LTRB,
        relative = False,
        image_size: "tuple[int, int]" = None, 
    ) -> "BoundingBox":
        if relative:
            assert image_size is not None, "For relative coordinates `image_size` should be provided."
            coords = cls.rel_to_abs(coords, image_size)

        if box_format is BoxFormat.LTWH:
            coords = cls.ltwh_to_ltrb(coords)
        elif box_format is BoxFormat.XYWH:
            coords = cls.xywh_to_ltrb(coords)
        elif box_format is BoxFormat.LTRB:
            pass
        else:
            raise ValueError(f"Unknown BoxFormat '{box_format}'.")

        return cls(label, *coords, confidence=confidence)

    @staticmethod
    def from_txt(
        string: str,
        box_format = BoxFormat.LTRB,
        relative = False,
        image_size: "tuple[int, int]" = None,
        separator: str = " "
    ) -> "BoundingBox":
        values = string.strip().split(separator)

        if len(values) == 5:
            label, *coords = values  
            confidence = None 
        elif len(values) == 6:
            label, confidence, *coords = values
        else:
            raise ParsingError(f"line '{string}' should have 5 or 6 values separated by whitespaces, not {len(values)}")

        try:
            coords = (float(c) for c in coords)
            if confidence is not None:
                confidence = float(confidence)
        except ValueError as e:
            raise ParsingError(f"{e} in line '{string}'")

        return BoundingBox.create(
            label=label, 
            coords=coords, 
            confidence=confidence, 
            box_format=box_format, 
            relative=relative, 
            image_size=image_size)

    @staticmethod
    def from_yolo(string: str, image_size: "tuple[int, int]") -> "BoundingBox":
        return BoundingBox.from_txt(string, 
            box_format=BoxFormat.XYWH, 
            relative=True, 
            image_size=image_size, 
            separator=" ")

    @staticmethod
    def from_xml(node: et.Element) -> "BoundingBox":
        try:
            label = node.findtext("name")
            box_node = node.find("bndbox")
            coords = (float(box_node.findtext(c)) 
                for c in ("xmin", "ymin", "xmax", "ymax"))
        except et.ParseError as e:
            line, _ = e.position
            raise ParsingError(f"syntax error at line {line}")
        except ValueError as e:
            raise ParsingError(f"{e}")

        return BoundingBox(label, *coords)

    @staticmethod
    def from_labelme(node: dict) -> "BoundingBox":
        # TODO: Add error handling
        # TODO: Handle if 'shape_type' is not rectangle
        label = str(node["label"])
        (xmin, ymin), (xmax, ymax) = node["points"]
        coords = (float(c) for c in (xmin, ymin, xmax, ymax))
        return BoundingBox(label, *coords)

    @staticmethod
    def from_cvat(node: et.Element) -> "BoundingBox":
        # TODO: Add error handling
        label = node.attrib["label"]
        coords = (float(node.attrib[c]) for c in ("xtl", "ytl", "xbr", "ybr"))
        return BoundingBox(label, *coords)

    def to_txt(self, 
        label_to_id: Mapping[str, Union[int, str]] = None,
        box_format: BoxFormat = BoxFormat.LTRB, 
        relative = False, 
        image_size: "tuple[int, int]" = None,
        separator: str = " "
    ) -> str:
        if box_format is BoxFormat.LTRB:
            coords = self.ltrb
        elif box_format is BoxFormat.XYWH:
            coords = self.xywh
        elif box_format is BoxFormat.LTWH:
            coords = self.ltwh
        else:
            raise ValueError(f"Unknown BoxFormat '{box_format}'")
        
        if relative:
            assert image_size is not None, "For relative coordinates, `image_size` should be provided."
            coords = BoundingBox.abs_to_rel(coords, image_size)

        assert "\n" not in separator, "The newline character '\\n' cannot be used as the separator character."

        label = self.label
        if label_to_id is not None:
            label = label_to_id[label]
        
        if isinstance(label, str):
            assert separator not in label, f"The box label '{label}' contains the character '{separator}' which is the same as the separtor character used for BoundingBox representation in TXT/YOLO format. This will corrupt the saved annotation file and likely make it unreadable. Use another character in the label name or `label_to_id` mapping, e.g. use and underscore instead of a whitespace."

        if self.is_ground_truth:
            return separator.join(f"{v}" for v in (label, *coords))
        else:
            return separator.join(f"{v}" for v in (label, self._confidence, *coords))

    def to_yolo(self,
        image_size: "tuple[int, int]",
        label_to_id: Mapping[str, Union[int, str]] = None
    ) -> str:
        return self.to_txt(
            label_to_id=label_to_id, 
            box_format=BoxFormat.XYWH, 
            relative=True, 
            image_size=image_size, 
            separator=" ")

    def to_labelme(self) -> dict:
        xmin, ymin, xmax, ymax = self.ltrb
        return {
            "label": self.label, 
            "points": [[xmin, ymin], [xmax, ymax]], 
            "shape_type": "rectangle"}

    def to_xml(self) -> et.Element:
        obj_node = et.Element("object")
        et.SubElement(obj_node, "name").text = self.label
        box_node = et.SubElement(obj_node, "bndbox")
        for tag, coord in zip(("xmin", "ymin", "xmax", "ymax"), self.ltrb):
            et.SubElement(box_node, tag).text = f"{coord}"
        return obj_node

    def to_cvat(self) -> et.Element:
        xtl, ytl, xbr, ybr = self.ltrb
        return et.Element("box", attrib={
            "label": self.label,
            "xtl": f"{xtl}",
            "ytl": f"{ytl}",
            "xbr": f"{xbr}",
            "ybr": f"{ybr}",
        })

    def __repr__(self) -> str:
        return f"BoundingBox(label: {self.label}, xmin: {self._xmin}, ymin: {self._ymin}, xmax: {self._xmax}, ymax: {self._ymax}, confidence: {self._confidence})"


from .boundingbox import BoundingBox, BoxFormat
from .annotation import Annotation
from .errors import UnknownImageFormat, FileParsingError
from .file_utils import glob
from .image_utils import get_image_size
from .atomic import open_atomic
from .thread_utils import thread_map

from typing import Dict, Callable, Iterator, Mapping, TypeVar, Iterable, Union
from csv import DictReader, DictWriter
from pathlib import Path
import xml.etree.ElementTree as et
from collections import defaultdict
import json
from tqdm import tqdm

T = TypeVar("T")
D = TypeVar("D")


class AnnotationSet:
    """Represents a set of annotations.
        
        This class efficiently stores annotations for several images (a validation 
        database for instance) with a fast lookup by `image_id`. It behaves like a 
        set or dictionary and has similar methods (contains, update, iterator, + 
        operator)."""

    def __init__(self, annotations: Iterable[Annotation] = None, override = False):
        # TODO: Add optional addition of labels found during
        # parsing, for instance COCO names and YOLO `.names`.
        # Could also add a (lazy) computed accessor that 
        # runs through all boxes to get labels.

        self._annotations: Dict[str, Annotation] = {}

        if annotations is not None:
            for annotation in annotations:
                self.add(annotation, override)

        self._id_to_label: "dict[int, str]" = None
        self._id_to_imageid: "dict[int, str]" = None

    def __getitem__(self, image_id: str) -> Annotation:
        return self._annotations[image_id]

    def get(self, image_id: str, default: D = None) -> Union[Annotation, D]:
        return self._annotations.get(image_id, default)

    def __len__(self) -> int:
        return len(self._annotations)

    def __iter__(self):
        yield from self._annotations.values()

    def items(self):
        return self._annotations.items()

    def __contains__(self, annotation: Annotation) -> bool:
        return annotation.image_id in self._annotations.keys()

    def add(self, annotation: Annotation, override = False):
        """Add an annotation to the set. The annotation 'image_id'
        should not already be present in the set.
        
        Parameters:
        - annotation: the annotation to add.
        - override: set to true to override any annotation with the same
        'image_id' already present in the set with the annotation."""

        if not override:
            assert annotation.image_id not in self.image_ids, \
                f"The annotation with id '{annotation.image_id}' is already present in the set (set `override` to True to remove this assertion)."
        self._annotations[annotation.image_id] = annotation

    def update(self, other: "AnnotationSet", override = False) -> "AnnotationSet":
        """Add annotations from another set to this set.
        
        Parameters:
        - other: the set of annotations to add.
        - override: set to true to override any annotation with the same
        'image_id' already present in the set.
        
        Returns:
        - the input set."""

        if not override:
            assert self.image_ids.isdisjoint(other.image_ids), \
                "some image ids are already in the set (set 'override' to True to remove this assertion)."
        self._annotations.update(other._annotations)
        return self

    def __ior__(self, other: "AnnotationSet") -> "AnnotationSet":
        return self.update(other)

    def __or__(self, other: "AnnotationSet") -> "AnnotationSet":
        return AnnotationSet().update(self).update(other)

    def map_labels(self, mapping: Mapping[str, str]) -> "AnnotationSet":
        for annotation in self:
            annotation.map_labels(mapping)
        return self

    @property
    def image_ids(self):
        return self._annotations.keys()

    @property
    def all_boxes(self) -> Iterator[BoundingBox]:
        """An iterator of all the bounding boxes."""
        for annotation in self:
            yield from annotation.boxes

    def nb_boxes(self) -> int:
        return sum(len(ann.boxes) for ann in self)

    def _labels(self) -> "set[str]":
        return {b.label for b in self.all_boxes}

    @staticmethod
    def from_iter(
        parser: Callable[[T], Annotation],
        iterable: Iterable[T], *,
        verbose: bool = False
    ) -> "AnnotationSet":
        annotations = thread_map(parser, iterable, desc="Parsing", verbose=verbose)
        return AnnotationSet(annotations)

    @staticmethod
    def from_folder(folder: Path, *,
        extension: str,
        parser: Callable[[Path], Annotation],
        recursive=False,
        verbose: bool = False
    ) -> "AnnotationSet":
        assert folder.is_dir()
        files = list(glob(folder, extension, recursive=recursive))
        return AnnotationSet.from_iter(parser, files, verbose=verbose)

    @staticmethod
    def from_txt(folder: Path, *,
        image_folder: Path = None,
        box_format = BoxFormat.LTRB,
        relative = False,
        file_extension: str = ".txt",
        image_extension: str = ".jpg",
        separator: str = " ",
        verbose: bool = False
    ) -> "AnnotationSet":
        """This method won't try to retreive the image sizes by default. Specify `image_folder` if you need them. `image_folder` is required when `relative` is True."""
        # TODO: Add error handling
        assert folder.is_dir()
        assert image_extension.startswith(".")

        if relative:
            assert image_folder is not None, "When `relative` is set to True, `image_folder` must be provided to read image sizes."
        if image_folder is not None:
            assert image_folder.is_dir()

        def _get_annotation(file: Path) -> Annotation:
            if image_folder is not None:
                image_path = image_folder / file.with_suffix(image_extension).name
                try:
                    image_size = get_image_size(image_path)
                except UnknownImageFormat:
                    raise FileParsingError(image_path, 
                        reason=f"Unable to read image file '{image_path}' to get the image size.")
            else:
                image_size = None
            
            return Annotation.from_txt(
                file_path=file,
                image_id=image_path.name,
                box_format=box_format,
                relative=relative,
                image_size=image_size,
                separator=separator)

        return AnnotationSet.from_folder(folder, 
            extension=file_extension, 
            parser=_get_annotation, 
            verbose=verbose)

    @staticmethod
    def from_yolo(folder: Path, *,
        image_folder: Path = None, 
        image_extension = ".jpg",
        verbose: bool = False
    ) -> "AnnotationSet":
        return AnnotationSet.from_txt(folder, 
            image_folder=image_folder, 
            box_format=BoxFormat.XYWH, 
            relative=True, 
            image_extension=image_extension,
            verbose=verbose)

    @staticmethod
    def from_xml(folder: Path, verbose: bool = False) -> "AnnotationSet":
        return AnnotationSet.from_folder(folder, 
            extension=".xml", 
            parser=Annotation.from_xml,
            verbose=verbose)      

    @staticmethod
    def from_openimage(file_path: Path, *, 
        image_folder: Path,
        verbose: bool = False
    ) -> "AnnotationSet":
        assert file_path.is_file() and file_path.suffix == ".csv", f"OpenImage annotation file {file_path} must be a csv file."
        assert image_folder.is_dir(), f"Image folder {image_folder} must be a valid directory."
        
        # TODO: Add error handling.
        annotations = AnnotationSet()

        with file_path.open(newline="") as f:
            reader = DictReader(f)

            for row in tqdm(reader, desc="Parsing", disable=not verbose):
                image_id = row["ImageID"]
                label = row["LabelName"]
                coords = (float(row[r]) for r in ("XMin", "YMin", "XMax", "YMax"))
                confidence = row.get("Confidence")
                
                if confidence is not None and confidence != "":
                    confidence = float(confidence)
                else:
                    confidence = None

                if image_id not in annotations.image_ids:
                    image_path = image_folder / image_id
                    image_size = get_image_size(image_path)
                    annotations.add(Annotation(image_id=image_id, image_size=image_size))
                
                annotation = annotations[image_id]
                annotation.add(
                    BoundingBox.create(
                        label=label, 
                        coords=coords, 
                        confidence=confidence, 
                        relative=True, 
                        image_size=annotation.image_size))

        return annotations

    @staticmethod
    def from_labelme(folder: Path, verbose: bool = False) -> "AnnotationSet":
        return AnnotationSet.from_folder(folder, 
            extension=".json", 
            parser=Annotation.from_labelme,
            verbose=verbose)

    @staticmethod
    def from_coco(file_path: Path, verbose: bool = False) -> "AnnotationSet":
        assert file_path.is_file() and file_path.suffix == ".json", f"COCO annotation file {file_path} must be a json file."

        # TODO: Add error handling
        with file_path.open() as f:
            content = json.load(f)

        id_to_label = {int(d["id"]): str(d["name"]) 
            for d in content["categories"]}
        id_to_annotation = {int(d["id"]): Annotation._from_coco_partial(d)
                for d in content["images"]}
        
        for element in tqdm(content["annotations"], desc="Parsing", disable=not verbose):
            annotation = id_to_annotation[int(element["image_id"])]
            label = id_to_label[int(element["category_id"])]
            coords = (float(c) for c in element["bbox"])
            confidence = element.get("score")
    
            if confidence is not None:
                confidence = float(confidence)

            annotation.add(BoundingBox.create(label=label, coords=coords, confidence=confidence, box_format=BoxFormat.LTWH))

        annotation_set = AnnotationSet(id_to_annotation.values())
        annotation_set._id_to_label = id_to_label
        annotation_set._id_to_imageid = {idx: ann.image_id for idx, ann in id_to_annotation.items()}

        return annotation_set

    def from_results(self, file_path: Path, verbose: bool = False) -> "AnnotationSet":
        assert file_path.is_file() and file_path.suffix == ".json", f"COCO annotation file {file_path} must be a json file."

        id_to_label = self._id_to_label
        id_to_imageid = self._id_to_imageid

        assert id_to_label is not None and id_to_imageid is not None, "The AnnotationSet instance should have been created with `AnnotationSet.from_coco()` or should have `self.id_to_label` and `self.id_to_image_id` populated. If not the case use the static method `AnnotationSet.from_coco_results()` instead."

        id_to_annotation = {}

        with file_path.open() as f:
            annotations = json.load(f)
        
        # TODO: Factorize this with `Self.from_coco()`?
        for element in tqdm(annotations, desc="Parsing", disable=not verbose):
            image_id = id_to_imageid[element["image_id"]]
            gt_ann = self[image_id]

            if image_id not in id_to_annotation:
                annotation = Annotation(image_id, (gt_ann.image_width, gt_ann.image_height))
                id_to_annotation[image_id] = annotation
            else:
                annotation = id_to_annotation[image_id]

            label = id_to_label[int(element["category_id"])]
            coords = (float(c) for c in element["bbox"])
            confidence = float(element["score"])

            annotation.add(BoundingBox.create(label=label, coords=coords, confidence=confidence, box_format=BoxFormat.LTWH))     

        annotation_set = AnnotationSet(id_to_annotation.values())
        annotation_set._id_to_label = id_to_label
        annotation_set._id_to_imageid = id_to_imageid
        return annotation_set

    @staticmethod
    def from_coco_results(file_path: Path, *,
        id_to_label: "dict[int, str]", 
        id_to_imageid: "dict[int, str]",
        verbose: bool = False
    ) -> "AnnotationSet":
        assert file_path.is_file() and file_path.suffix == ".json", f"COCO annotation file {file_path} must be a json file."

        id_to_annotation = {}

        with file_path.open() as f:
            annotations = json.load(f)
        
        # TODO: Factorize this with `Self.from_coco()`?
        for element in tqdm(annotations, desc="Parsing", disable=not verbose):
            image_id = id_to_imageid[element["image_id"]]

            if image_id not in id_to_annotation:
                annotation = Annotation(image_id)
                id_to_annotation[image_id] = annotation
            else:
                annotation = id_to_annotation[image_id]

            label = id_to_label[int(element["category_id"])]
            coords = (float(c) for c in element["bbox"])
            confidence = float(element["score"])

            annotation.add(BoundingBox.create(label=label, coords=coords, confidence=confidence, box_format=BoxFormat.LTWH))      

        annotation_set = AnnotationSet(id_to_annotation.values())
        annotation_set._id_to_label = id_to_label
        annotation_set._id_to_imageid = id_to_imageid
        return annotation_set

    @staticmethod
    def from_cvat(file_path: Path, verbose: bool = False) -> "AnnotationSet":
        assert file_path.is_file() and file_path.suffix == ".xml", f"CVAT annotation file {file_path} must be a xml file."
        
        # TODO: Add error handling.
        with file_path.open() as f:
            root = et.parse(f).getroot()
        image_nodes = list(root.iter("image"))
        return AnnotationSet.from_iter(Annotation.from_cvat, image_nodes, verbose=verbose)

    def save_from_it(self, save_fn: Callable[[Annotation], None], *, verbose: bool = False):        
        thread_map(save_fn, self, desc="Saving", verbose=verbose)

    def save_txt(self, save_dir: Path, *,
        label_to_id: Mapping[str, Union[int, str]] = None,
        box_format: BoxFormat = BoxFormat.LTRB, 
        relative: bool = False, 
        separator: str = " ",
        file_extension: str = ".txt",
        verbose: bool = False
    ):
        save_dir.mkdir(exist_ok=True)

        def _save(annotation: Annotation):
            image_id = annotation.image_id
            path = (save_dir / image_id).with_suffix(file_extension)
            annotation.save_txt(path, 
                label_to_id=label_to_id, 
                box_format=box_format, 
                relative=relative,
                separator=separator)

        self.save_from_it(_save, verbose=verbose)

    def save_yolo(self, save_dir: Path, *, 
        label_to_id: Mapping[str, Union[int, str]] = None,
        verbose: bool = False
    ):
        save_dir.mkdir(exist_ok=True)

        def _save(annotation: Annotation):
            path = save_dir / Path(annotation.image_id).with_suffix(".txt")
            annotation.save_yolo(path, label_to_id=label_to_id)

        self.save_from_it(_save, verbose=verbose)

    def save_labelme(self, save_dir: Path, *, verbose: bool = False):
        save_dir.mkdir(exist_ok=True)
        
        def _save(annotation: Annotation):
            path = save_dir / Path(annotation.image_id).with_suffix(".json")
            annotation.save_labelme(path)

        self.save_from_it(_save, verbose=verbose)

    def save_xml(self, save_dir: Path, *, verbose: bool = False):
        save_dir.mkdir(exist_ok=True)
        
        def _save(annotation: Annotation):
            path = save_dir / Path(annotation.image_id).with_suffix(".xml")
            annotation.save_xml(path)

        self.save_from_it(_save, verbose=verbose)

    def to_coco(self, *,
        label_to_id: "dict[str, int]" = None, 
        imageid_to_id: "dict[str, int]" = None,
        auto_ids: bool = False,
        verbose: bool = False
    ) -> dict:
        native_ids = (label_to_id or self._id_to_label) and (imageid_to_id or self._id_to_imageid)
        assert native_ids or auto_ids, "For COCO, mappings from labels and image ids to integer ids are required. They can be provided either by argument or automatically by the `AnnotationSet` instance if it was created with `AnnotationSet.from_coco()` or `AnnotationSet.from_coco_results()`. You can also set `auto_ids` to True to automatically create image and label ids (warning: this could cause unexpected compatibility issues with other COCO datasets)."

        if native_ids:  # native_ids takes precedence over auto_ids
            label_to_id = label_to_id or {v: k for k, v in self._id_to_label.items()}
            imageid_to_id = imageid_to_id or {v: k for k, v in self._id_to_imageid.items()}
        else:
            label_to_id = {l: i for i, l in enumerate(sorted(self._labels()))}
            imageid_to_id = {im: i for i, im in enumerate(sorted(self.image_ids))}

        annotations = []
        ann_id_count = 0
        for annotation in tqdm(self, desc="Saving", disable=not verbose):
            for box in annotation.boxes:
                box_annotation = {
                    "iscrowd": 0, "ignore": 0,
                    "image_id": imageid_to_id[annotation.image_id],
                    "bbox": box.ltwh,
                    "category_id": label_to_id[box.label],
                    "id": ann_id_count}

                if box.is_detection:
                    box_annotation["score"] = box.confidence

                annotations.append(box_annotation)

                ann_id_count += 1

        images = [{
            "id": imageid_to_id[a.image_id], 
            "file_name": a.image_id, 
            "width": a.image_width, 
            "height": a.image_height} for a in self]

        categories = [{"supercategory": "none", "id": i, "name": l} for l, i in label_to_id.items()]

        return {"images": images, "annotations": annotations, "categories": categories}

    def save_coco(self, path: Path, *,
        label_to_id: "dict[str, int]" = None, 
        imageid_to_id: "dict[str, int]" = None,
        auto_ids: bool = False,
        verbose: bool = False
    ):
        if path.suffix == "":
            path = path.with_suffix(".json")

        assert path.suffix == ".json"
        
        content = self.to_coco(
            label_to_id=label_to_id, 
            imageid_to_id=imageid_to_id, 
            auto_ids=auto_ids, 
            verbose=verbose)

        with open_atomic(path, "w") as f:
            json.dump(content, fp=f, allow_nan=False)

    def save_openimage(self, path: Path, *, verbose: bool = False):
        if path.suffix == "":
            path = path.with_suffix(".csv")

        assert path.suffix == ".csv"

        fields = (
            "ImageID", "Source", "LabelName", "Confidence", 
            "XMin", "XMax", "YMin", "YMax", "IsOccluded", 
            "IsTruncated", "IsGroupOf", "IsDepiction", "IsInside")
    
        with open_atomic(path, "w", newline="") as f:
            writer = DictWriter(f, fieldnames=fields, restval="")
            writer.writeheader()

            for annotation in tqdm(self, desc="Saving", disable=not verbose):
                for box in annotation.boxes:
                    label = box.label
                    
                    assert "," not in label, f"The box label '{label}' contains the character ',' which is the same as the separtor character used for BoundingBox representation in OpenImage format (CSV). This will corrupt the saved annotation file and likely make it unreadable. Use another character in the label name, e.g. use and underscore instead of a comma."

                    xmin, ymin, xmax, ymax = BoundingBox.abs_to_rel(coords=box.ltrb, size=annotation.image_size)
                    
                    row = {
                        "ImageID": annotation.image_id,
                        "LabelName": label,
                        "XMin": xmin, "XMax": xmax, "YMin": ymin, "YMax": ymax}
                    
                    if box.is_detection:
                        row["Confidence"] = box.confidence
                    
                    writer.writerow(row)

    def to_cvat(self, *, verbose: bool = False) -> et.Element:
        def _create_node(annotation: Annotation) -> et.Element:
            return annotation.to_cvat()

        sub_nodes = thread_map(_create_node, self, desc="Saving", verbose=verbose)
        node = et.Element("annotations")
        node.extend(sub_nodes)
        return node

    def save_cvat(self, path: Path, *, verbose: bool = False):
        if path.suffix == "":
            path = path.with_suffix(".xml")
        assert path.suffix == ".xml"
        content = self.to_cvat(verbose=verbose)
        content = et.tostring(content, encoding="unicode")
        with open_atomic(path, "w") as f:
            f.write(content)

    @staticmethod
    def parse_names_file(path: Path) -> "dict[str, str]":
        """Parse .names file.
        
        Parameters:
        - path: the path to the .names file.

        Returns: 
        - A dictionary mapping label number with label names."""
        # TODO: Add error handling
        return {str(i): v for i, v in enumerate(path.read_text().splitlines())}

    @staticmethod
    def parse_mid_file(path: Path) -> "dict[str, str]":
        with path.open() as f:
            reader = DictReader(f, fieldnames=("LabelName", "DisplayName"))
            return {l["LabelName"]: l["DisplayName"] for l in reader}

    def show_stats(self, *, verbose: bool = False):
        from rich.table import Table
        from rich import print as rprint

        box_by_label = defaultdict(int)
        im_by_label = defaultdict(int)

        for annotation in tqdm(self, desc="Stats", disable=not verbose):
            for box in annotation.boxes:
                box_by_label[box.label] += 1
            
            labels = annotation._labels()

            for label in labels:
                im_by_label[label] += 1
            
            if len(labels) == 0:
                im_by_label["<empty_image>"] += 1
            
        tot_box = sum(box_by_label.values())
        tot_im = len(self)

        table = Table(title="Database Stats", show_footer=True)
        table.add_column("Label", footer="Total")
        table.add_column("Images", footer=f"{tot_im}", justify="right")
        table.add_column("Boxes", footer=f"{tot_box}", justify="right")

        for label in sorted(im_by_label.keys()):
            nb_im = im_by_label[label]
            nb_box = box_by_label[label]
            table.add_row(label, f"{nb_im}", f"{nb_box}")

        rprint(table)

    def __repr__(self) -> str:
        return f"AnnotationSet(annotations: {self._annotations})"


from .boundingbox import BoxFormat, Coordinates, BoundingBox
from .annotation import Annotation
from .annotationset import AnnotationSet
from .errors import UnknownImageFormat, ParsingError, FileParsingError
from .evalutation import EvaluationItem, Evaluation, COCOEvaluator


from pathlib import Path
from typing import Union, Iterable


def glob(
    folder: Path, 
    extensions: Union[str, Iterable[str]], 
    recursive: bool = False
) -> Iterable[Path]:
    """Glob files by providing extensions to match."""
    if isinstance(extensions, str):
        extensions = {extensions}
    else:
        extensions = set(extensions)

    assert all(e.startswith(".") for e in extensions), \
        "Parameter `extension` should start with a dot."

    files = folder.glob("**/*") if recursive else folder.glob("*")

    return (f for f in files \
        if f.suffix in extensions and not f.name.startswith("."))


from .boundingbox import BoxFormat
from .annotationset import AnnotationSet
from .evalutation import COCOEvaluator

import argparse
from pathlib import Path


PARSE_CHOICES = {"coco", "yolo", "labelme", "pascalvoc", "openimage", "txt", "cvat"}
PARSE_CHOICES_EXT = {*PARSE_CHOICES, "coco_result"}


def parse_args():
    parser = argparse.ArgumentParser()
    
    parser.add_argument("--quiet", "-q", action="store_true")
    parser.add_argument("--threads", "-j", default=None, type=int)

    subparsers = parser.add_subparsers(dest="mode")
    convert_parser = subparsers.add_parser("convert")
    stats_parser = subparsers.add_parser("summary")
    eval_parser = subparsers.add_parser("evaluate")

    add_convert_args(convert_parser)
    add_stats_args(stats_parser)
    add_eval_args(eval_parser)

    return parser.parse_args()


def add_parse_args(parser: argparse.ArgumentParser, metavar: str = "input", label: str = None):
    parser.add_argument("input", type=Path, metavar=metavar)
    
    group = parser.add_argument_group("Parse options" if label is None else label)
    group.add_argument("--format", "-f", type=str, choices=PARSE_CHOICES, dest="format_in")
    group.add_argument("--img_folder", '-d', type=Path, default=None)
    group.add_argument("--mapping", "-m", type=Path, default=None, dest="mapping_in")
    group.add_argument("--bb_fmt", "-b", type=str, choices=("ltrb", "ltwh", "xywh"), default="ltrb", dest="bb_fmt_in")
    group.add_argument("--norm", "-n", type=str, choices=("abs", "rel"), default="abs", dest="norm_in")
    group.add_argument("--ext", "-e", type=str, default=".txt", dest="ext_in")
    group.add_argument("--img_ext", "-g", type=str, default=".jpg", dest="img_ext_in")
    group.add_argument("--sep", "-p", type=str, default=" ", dest="sep_in")


def add_parse_dets_args(parser: argparse.ArgumentParser):
    parser.add_argument("predictions", type=Path)

    group = parser.add_argument_group("Predictions parse options")
    group.add_argument("--format_dets", "-F", type=str, choices=PARSE_CHOICES_EXT)
    group.add_argument("--img_folder_dets", '-D', type=Path, default=None)
    group.add_argument("--mapping_dets", "-M", type=Path, default=None)
    group.add_argument("--bb_fmt_dets", "-B", type=str, choices=("ltrb", "ltwh", "xywh"), default="ltrb", dest="bb_fmt_dets")
    group.add_argument("--norm_dets", "-N", type=str, choices=("abs", "rel"), default="abs", dest="norm_in_dets")
    group.add_argument("--ext_dets", "-E", type=str, default=".txt")
    group.add_argument("--img_ext_dets", "-G", type=str, default=".jpg")
    group.add_argument("--sep_dets", "-P", type=str, default=" ")


def add_save_args(parser: argparse.ArgumentParser):
    parser.add_argument("output", type=Path)

    group = parser.add_argument_group("Save options")
    group.add_argument("--save_fmt", "-F", type=str, choices=PARSE_CHOICES, dest="format_out")  # TODO: add PARSE_CHOICES_EXT
    group.add_argument("--bb_fmt_out", "-B", type=str, choices=("ltrb", "ltwh", "xywh"), default="ltrb")
    group.add_argument("--norm_out", "-N", type=str, choices=("abs", "rel"), default="abs")
    group.add_argument("--sep_out", "-P", type=str, default=" ")
    group.add_argument("--ext_out", "-E", type=str, default=" ")
    group.add_argument("--coco_auto_ids", "-A", action="store_true")

    mapping_group = group.add_mutually_exclusive_group()
    mapping_group.add_argument("--mapping_out", "-M", type=Path, default=None)
    mapping_group.add_argument("--reverse_mapping_out", "-R", type=Path, default=None)


def add_stats_args(parser: argparse.ArgumentParser):
    add_parse_args(parser)


def add_convert_args(parser: argparse.ArgumentParser): 
    add_parse_args(parser)
    add_save_args(parser)


def add_eval_args(parser: argparse.ArgumentParser):
    add_parse_args(parser, metavar="groundtruths", label="Ground-truths parse options")
    add_parse_dets_args(parser)
    
    parser.add_argument("--save", "-s", type=Path, default=None, dest="save_csv_path")

    # parser.add_argument("--ap", action="append", dest="metrics")
    # parser.add_argument("--ap50", action="append", dest="metrics")
    # parser.add_argument("--iou", type=int, default=None)  # mutually_exclusive_group()
    # etc...


def parse_annotations(args: argparse.Namespace) -> AnnotationSet:
    input: Path = args.input.expanduser().resolve()
    format_in: str = args.format_in
    verbose: bool = not args.quiet

    if format_in == "coco":
        return AnnotationSet.from_coco(input, verbose=verbose)
    elif format_in == "pascalvoc":
        return AnnotationSet.from_xml(input, verbose=verbose)
    elif format_in == "openimage":
        assert args.img_folder is not None, "The image directory must be provided for openimage format (required for reading the image size)."
        image_dir: Path = args.img_folder.expanduser().resolve()
        return AnnotationSet.from_openimage(input, image_folder=image_dir, verbose=verbose)
    elif format_in == "labelme":
        return AnnotationSet.from_labelme(input, verbose=verbose)
    elif format_in == "cvat":
        return AnnotationSet.from_cvat(input, verbose=verbose)
    else:
        img_ext: str = args.img_ext_in
        if args.img_folder is not None:
            image_dir: Path = args.img_folder.expanduser().resolve()
        else:
            image_dir = None

        if format_in == "yolo":
            annotations = AnnotationSet.from_yolo(input, 
                image_folder=image_dir, 
                image_extension=img_ext, 
                verbose=verbose)    
        elif format_in == "txt":
            format_in = BoxFormat.from_string(args.bb_fmt_in)
            relative: bool = args.norm_in == "rel"
            extension: str = args.ext_in
            sep: str = args.sep_in

            annotations = AnnotationSet.from_txt(input, 
                image_folder=image_dir, 
                box_format=format_in, 
                relative=relative,
                file_extension=extension, 
                image_extension=img_ext, 
                separtor=sep, 
                verbose=verbose)
        else:
            raise ValueError(f"Input format '{format_in}' unknown")

        if args.mapping_in is not None:
            map_path: Path = args.mapping_in.expanduser().resolve()
            mapping = AnnotationSet.parse_names_file(map_path)
            annotations.map_labels(mapping)
        
        return annotations


def parse_dets_annotations(args: argparse.Namespace, coco_gts: AnnotationSet = None) -> AnnotationSet:
    input: Path = args.predictions.expanduser().resolve()
    format_dets: str = args.format_dets
    verbose: bool = not args.quiet

    if format_dets == "coco":
        return AnnotationSet.from_coco(input, verbose=verbose)
    if format_dets == "coco_result":
        if coco_gts is None:
            raise ValueError("When using 'COCO results', the parsed ground truths must be in 'COCO' format.")
        return coco_gts.from_results(input, verbose=verbose)
    elif format_dets == "pascalvoc":
        return AnnotationSet.from_xml(input, verbose=verbose)
    elif format_dets == "openimage":
        assert args.image_dir_dets is not None, "The image directory must be provided for openimage format (required for reading the image size)."
        image_dir: Path = args.img_folder_dets.expanduser().resolve()
        return AnnotationSet.from_openimage(input, image_folder=image_dir, verbose=verbose)
    elif format_dets == "labelme":
        return AnnotationSet.from_labelme(input, verbose=verbose)
    elif format_dets == "cvat":
        return AnnotationSet.from_cvat(input, verbose=verbose)
    else:
        img_ext: str = args.img_ext_dets
        if args.image_dir is not None:
            image_dir: Path = args.img_folder_dets.expanduser().resolve()
        else:
            image_dir = None

        if format_dets == "yolo":
            annotations = AnnotationSet.from_yolo(input, 
                image_folder=image_dir, 
                image_extension=img_ext, 
                verbose=verbose)
        elif format_dets == "txt":
            bb_fmt = BoxFormat.from_string(args.bb_fmt_dets)
            relative: bool = args.norm_dets == "rel"
            extension: str = args.ext_dets
            sep: str = args.sep_dets

            annotations = AnnotationSet.from_txt(input, 
                image_folder=image_dir, 
                box_format=bb_fmt, 
                relative=relative, 
                file_extension=extension,
                image_extension=img_ext,
                separator=sep,
                verbose=verbose)
        else:
            raise ValueError(f"Groundtruth format '{format_dets}' unknown")

        if args.mapping_in is not None:
            map_path: Path = args.mapping_in.expanduser().resolve()
            mapping = AnnotationSet.parse_names_file(map_path)
            annotations.map_labels(mapping)

        return annotations


# TODO: Add "coco_result"
def save_annotations(args: argparse.Namespace, annotations: AnnotationSet):
    output: Path = args.output.expanduser().resolve()
    format_out: str = args.format_out
    verbose: bool = not args.quiet

    if args.mapping_out is not None:  # Takes precedence
        map_path: Path = args.mapping_out.expanduser().resolve()
        mapping = AnnotationSet.parse_names_file(map_path)
        annotations.map_labels(mapping)
    elif args.reverse_mapping_out is not None:
        map_path: Path = args.reverse_mapping_out.expanduser().resolve()
        mapping = AnnotationSet.parse_names_file(map_path)
        mapping = {v: k for k, v in mapping.items()}
        annotations.map_labels(mapping)

    if format_out == "coco":
        annotations.save_coco(output, auto_ids=args.coco_auto_ids, verbose=verbose)
    elif format_out == "pascalvoc":
        annotations.save_xml(output, verbose=verbose)
    elif format_out == "openimage":
        annotations.save_openimage(output, verbose=verbose)
    elif format_out == "labelme":
        annotations.save_labelme(output, verbose=verbose)
    elif format_out == "cvat":
        annotations.save_cvat(output, verbose=verbose)
    else:
        if format_out == "yolo":
            annotations.save_yolo(output, verbose=verbose)
        elif format_out == "txt":
            bb_fmt = BoxFormat.from_string(args.bb_fmt_out)
            relative: bool = args.norm_out == "rel"
            extension: str = args.ext_out
            sep: str = args.sep_out

            annotations.save_txt(output, 
                label_to_id=None, 
                box_format=bb_fmt, 
                relative=relative, 
                separator=sep, 
                file_extension=extension, 
                verbose=verbose)
        else:
            raise ValueError(f"Save format '{format_out}' unknown")


def evaluate(args: argparse.Namespace, groundtruths: AnnotationSet, predictions: AnnotationSet):
    verbose: bool = not args.quiet
    evaluator = COCOEvaluator(groundtruths, predictions)
    evaluator.show_summary(verbose=verbose)
    
    if args.save_csv_path is not None:
        path = args.save_csv_path.expanduser().resolve()
        evaluator.save_csv(path, verbose=False)
        if verbose:
            print(f"Evaluation saved to '{args.save_csv_path}'.")


def main():
    args = parse_args()   

    assert args.threads is None or args.threads > 0

    mode = args.mode
    if mode == "convert":
        annotations = parse_annotations(args)
        save_annotations(args, annotations)
    elif mode == "summary":
        annotations = parse_annotations(args)
        annotations.show_stats()
    elif mode == "evaluate":
        groundtruths = parse_annotations(args)
        coco_gts = groundtruths if args.format_dets == "coco_result" else None
        predictions = parse_dets_annotations(args, coco_gts=coco_gts)
        evaluate(args, groundtruths, predictions)
    else:
        raise ValueError(f"Sub-command '{mode}' not recognized.")


if __name__ == "__main__":
    main()


from pathlib import Path
from setuptools import setup, find_packages


NAME = "globox"
AUTHOR = "Louis Lac"
EMAIL = "lac.louis5@gmail.com"
URL = "https://github.com/laclouis5/globox"
DESCRIPTION = "Globox is a package and command line interface to read and convert object detection databases (COCO, YOLO, PascalVOC, LabelMe, CVAT, OpenImage, ...) and evaluate them with COCO and PascalVOC."

LICENSE = "MIT"
PYTHON = ">=3.7"

REQUIREMENTS = ["rich", "tqdm", "numpy"]
EXTRA_REQ = ["tox", "pytest", "twine", "build", "pycocotools"]

with open("README.md") as f:
    LONG_DESCRIPTION = f.read()

with (Path("src/", NAME) / "__version__.py").open() as f:
    about = {}
    exec(f.read(), about)
    VERSION = about["__version__"]

setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    author=AUTHOR,
    author_email=EMAIL,
    url=URL,
    license=LICENSE,
    packages=find_packages(where="src/"),
    package_dir={"": "src"},
    python_requires=PYTHON,
    install_requires=REQUIREMENTS,
    extras_require={
        "dev": EXTRA_REQ
    },
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    keywords=["annotation", "metrics", "object detection", "bounding boxes", 
        "yolo", "openimages", "cvat", "coco", "pascal voc", "average precision", 
        "mean average precision",
    ],
    entry_points={
        "console_scripts": [f"{NAME}={NAME}.cli:main"],
    }
)


from .boundingbox import BoundingBox, BoxFormat
from .errors import ParsingError, FileParsingError
from .atomic import open_atomic

from pathlib import Path
from typing import Mapping, Optional, Union
import xml.etree.ElementTree as et
import json


class Annotation: 
    """Stores bounding box annotations for one image.

    The image should be uniquely identified by the `image_id` str. The
    image size is necessary for some operations."""

    def __init__(self, 
        image_id: str, 
        image_size: Optional["tuple[int, int]"] = None, 
        boxes: "list[BoundingBox]" = None
    ) -> None:
        if image_size is not None:
            img_w, img_h = image_size
            assert img_w > 0 and img_h > 0
            assert int(img_w) == img_w and int(img_h) == img_h

        self.image_id = image_id
        self.image_size = image_size
        self.boxes = boxes or []

    @property
    def image_width(self) -> Optional[int]:
        return self.image_size[0]

    @property
    def image_height(self) -> Optional[int]:
        return self.image_size[1]

    def add(self, box: BoundingBox):
        self.boxes.append(box)

    def map_labels(self, mapping: Mapping[str, str]) -> "Annotation":
        """Change all the bounding box annotation labels according to
        the provided mapping (act as a translation)."""
        for box in self.boxes:
            box.label = mapping[box.label]
        return self

    def _labels(self) -> "set[str]":
        return {b.label for b in self.boxes}

    @staticmethod
    def from_txt(
        file_path: Path,
        image_id: str,
        box_format: BoxFormat = BoxFormat.LTRB,
        relative = False,
        image_size: "tuple[int, int]" = None,
        separator: str = " "
    ) -> "Annotation":
        try:
            lines = file_path.read_text().splitlines()
        except OSError:
            raise FileParsingError(file_path, reason="cannot read file")

        try:
            boxes = [BoundingBox.from_txt(
                l, box_format, relative, image_size, separator)
                for l in lines]
        except ParsingError as e:
            raise FileParsingError(file_path, e.reason)

        return Annotation(image_id, image_size, boxes)

    @staticmethod
    def from_yolo(
        file_path: Path,
        image_id: str,
        image_size: "tuple[int, int]",
    ):
        return Annotation.from_txt(file_path, 
            image_id=image_id,
            box_format=BoxFormat.XYWH, 
            relative=True, 
            image_size=image_size, 
            separator=" ")

    @staticmethod
    def from_xml(file_path: Path) -> "Annotation":
        try:
            with file_path.open() as f:
                root = et.parse(f).getroot()
            
            image_id = root.findtext("filename")
            size_node = root.find("size")
            width = size_node.findtext("width")
            height = size_node.findtext("height")
            image_size = int(width), int(height) 
            boxes = [BoundingBox.from_xml(n) for n in root.iter("object")]
        except OSError:
            raise FileParsingError(file_path, reason="cannot read file")
        except et.ParseError as e:
            line, _ = e.position
            raise FileParsingError(file_path, reason=f"syntax error at line {line}")
        except ParsingError as e:
            raise FileParsingError(file_path, reason=e.reason)
        except ValueError as e:
            raise ParsingError(f"{e}")

        return Annotation(image_id, image_size, boxes)
        
    @staticmethod
    def from_labelme(file_path: Path) -> "Annotation":
        # TODO: Add error handling.
        with file_path.open() as f:
            content = json.load(f)
            if "imageData" in content: 
                del content["imageData"]

        image_id = str(content["imagePath"])
        width = int(content["imageWidth"])
        height = int(content["imageHeight"])
        boxes = [BoundingBox.from_labelme(n) for n in content["shapes"]
            if n["shape_type"] == "rectangle"]
        
        return Annotation(image_id, (width, height), boxes)

    @staticmethod
    def _from_coco_partial(node: dict) -> "Annotation":
        # TODO: Add error handling
        image_id = str(node["file_name"])
        image_size = int(node["width"]), int(node["height"])
        return Annotation(image_id, image_size)

    @staticmethod
    def from_cvat(node: et.Element) -> "Annotation":
        # TODO: Add error handling
        attribs = node.attrib
        image_id = attribs["name"]
        image_size = int(attribs["width"]), int(attribs["height"])
        boxes = [BoundingBox.from_cvat(n) for n in node.iter("box")]
        return Annotation(image_id, image_size, boxes)

    def to_txt(self, *,
        label_to_id: Mapping[str, Union[int, str]] = None,
        box_format: BoxFormat = BoxFormat.LTRB, 
        relative = False,
        image_size: "tuple[int, int]" = None, 
        separator: str = " "
    ) -> str:
        image_size = image_size or self.image_size
        return "\n".join(
            b.to_txt(label_to_id, box_format, relative, image_size, separator) for b in self.boxes)

    def to_yolo(self, *,
        label_to_id: Mapping[str, Union[int, str]] = None,
        image_size: "tuple[int, int]" = None
    ) -> str:
        image_size = image_size or self.image_size
        return "\n".join(b.to_yolo(image_size, label_to_id) for b in self.boxes)

    def save_txt(self, path: Path, *,
        label_to_id: Mapping[str, Union[int, str]] = None,
        box_format: BoxFormat = BoxFormat.LTRB, 
        relative = False, 
        image_size: "tuple[int, int]" = None,
        separator: str = " "
    ):
        content = self.to_txt(
            label_to_id=label_to_id, 
            box_format=box_format, 
            relative=relative, 
            image_size=image_size, 
            separator=separator)

        with open_atomic(path, "w") as f:
            f.write(content)

    def save_yolo(self, path: Path, *,
        label_to_id: Mapping[str, Union[int, str]] = None,
        image_size: "tuple[int, int]" = None
    ):
        content = self.to_yolo(label_to_id=label_to_id, image_size=image_size)
        
        with open_atomic(path, "w") as f:
            f.write(content)

    def to_labelme(self, *, image_size: "tuple[int, int]" = None) -> dict:
        image_size = image_size or self.image_size
        assert image_size is not None, "An image size should be provided either by argument or by `self.image_size`."

        return {
            "imagePath": self.image_id,
            "imageWidth": image_size[0],
            "imageHeight": image_size[1],
            "imageData": None,
            "shapes": [b.to_labelme() for b in self.boxes]}

    def save_labelme(self, path: Path, *, image_size: "tuple[int, int]" = None):
        content = self.to_labelme(image_size=image_size)
        with open_atomic(path, "w") as f:
            json.dump(content, fp=f, allow_nan=False)

    def to_xml(self, *, image_size: "tuple[int, int]" = None) -> et.Element:
        image_size = image_size or self.image_size
        assert image_size is not None, "An image size should be provided either by argument or by `self.image_size`."

        ann_node = et.Element("annotation")
        et.SubElement(ann_node, "filename").text = self.image_id

        size_node = et.SubElement(ann_node, "size")
        et.SubElement(size_node, "width").text = f"{image_size[0]}"
        et.SubElement(size_node, "height").text = f"{image_size[1]}"

        for box in self.boxes:
            ann_node.append(box.to_xml())

        return ann_node

    def save_xml(self, path: Path, *, image_size: "tuple[int, int]" = None):
        content = self.to_xml(image_size=image_size)
        content = et.tostring(content, encoding="unicode")
        
        with open_atomic(path, "w") as f:
            f.write(content)

    def to_cvat(self, *, image_size: "tuple[int, int]" = None) -> et.Element:
        image_size = image_size or self.image_size
        assert image_size is not None, "An image size should be provided either by argument or by `self.image_size`."

        img_node = et.Element("image", attrib={
            "name": self.image_id,
            "width": f"{image_size[0]}",
            "height": f"{image_size[1]}",
        })

        img_node.extend(box.to_cvat() for box in self.boxes)

        return img_node

    def __repr__(self) -> str:
        return f"Annotation(image_id: {self.image_id}, image_size: {self.image_size}, boxes: {self.boxes})"